#include "nssync.h"
#include <inttypes.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <tcpd.h>
#define MHD_PLATFORM_H
#include <microhttpd.h>

static int
setup_listen_socket(struct grecs_sockaddr *sa, int fd)
{
	if (bind(fd, sa->sa, sa->len) == -1) {
		error("%s: cannot bind: %s",
		      grecs_sockaddr_str(sa), strerror(errno));
		return -1;
	}

	if (listen(fd, 8) == -1) {
		error("%s: listen: %s",
		      grecs_sockaddr_str(sa), strerror(errno));
		return -1;
	}
	return 0;
}

static int
open_socket(struct grecs_sockaddr *sa)
{
	int reuse;
	int fd;

	fd = socket(sa->sa->sa_family, SOCK_STREAM, 0);
	if (fd == -1) {
		error("socket: %s", strerror(errno));
		return -1;
	}
	reuse = 1;
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
		   &reuse, sizeof(reuse));

	if (setup_listen_socket(sa, fd)) {
		close(fd);
		return -1;
	}
	return fd;
}

static void
nssync_mhd_panic(void *cls, const char *file, unsigned int line,
		 const char *reason)
{
	if (reason)
		error("%s:%d: MHD PANIC: %s", file, line, reason);
	else
		error("%s:%d: MHD PANIC", file, line);
	abort();
}

static void
nssync_mhd_logger(void *arg, const char *fmt, va_list ap)
{
	verror(fmt, ap);
}

static int
nssync_mhd_acl(void *cls, const struct sockaddr *addr,  socklen_t addrlen)
{
	struct request_info req;
	request_init(&req,
		     RQ_DAEMON, "nssync",
		     RQ_CLIENT_SIN, addr,
		     NULL);
	fromhost(&req);
	return hosts_access(&req) ? MHD_YES : MHD_NO;
}

static int
cb_not_found(void *cls,
	     struct MHD_Connection *connection,
	     const char *url, const char *method,
	     const char *version,
	     const char *upload_data, size_t *upload_data_size,
	     void **con_cls)
{
	int ret;
	struct MHD_Response *response;
	static char errtext[] = "Not found";
	
	response = MHD_create_response_from_buffer (strlen(errtext),
						    (void *) errtext,
						    MHD_RESPMEM_PERSISTENT);
	ret = MHD_queue_response(connection, MHD_HTTP_NOT_FOUND, response);
	MHD_destroy_response(response);
	return ret;
}

static int
cb_bad_method(void *cls,
	      struct MHD_Connection *connection,
	      const char *url, const char *method,
	      const char *version,
	      const char *upload_data, size_t *upload_data_size,
	      void **con_cls)
{
	int ret;
	struct MHD_Response *response;
	
	response = MHD_create_response_from_buffer (0,
						    "",
						    MHD_RESPMEM_PERSISTENT);
	ret = MHD_queue_response(connection, MHD_HTTP_METHOD_NOT_ALLOWED, response);
	MHD_destroy_response(response);
	return ret;
}

static void
json_writer(void *closure, char const *text, size_t len)
{
	grecs_txtacc_t acc = closure;
	grecs_txtacc_grow(acc, text, len);
}

static int
do_sync(void *cls,
	struct MHD_Connection *connection,
	const char *url, const char *method,
	const char *version,
	const char *upload_data, size_t *upload_data_size,
	void **con_cls)
{
	int ret, rc;
	struct MHD_Response *response;
	struct json_value *json;
	struct json_format fmt;
	grecs_txtacc_t acc;
	char *buf;
	
	if (*url)
		return cb_not_found(cls, connection, url, method,
				    version, upload_data, upload_data_size,
				    con_cls);

	if (strcmp(method, "POST"))
		return cb_bad_method(cls, connection, url, method,
				     version, upload_data, upload_data_size,
				     con_cls);
		
	rc = nssync(&json);

	acc = grecs_txtacc_create();
	fmt.indent = 0;
	fmt.precision = 0;
	fmt.write = json_writer;
	fmt.data = acc;
	json_format_value(json, &fmt);
	json_value_free(json);
	
	grecs_txtacc_grow_char(acc, 0);
	buf = grecs_txtacc_finish(acc, 1);
	grecs_txtacc_free(acc);
	
	response = MHD_create_response_from_buffer (strlen(buf),
						    (void *) buf,
						    MHD_RESPMEM_MUST_FREE);
	MHD_add_response_header(response, "Content-Type", "application/json");

	ret = MHD_queue_response(connection, MHD_HTTP_OK, response);
	MHD_destroy_response(response);
	return ret;
}


struct nssync_resource {
	char const *uri;
	size_t uri_len;
	MHD_AccessHandlerCallback cbf;
	int (*predicate)(void);
};

static struct nssync_resource nssync_resources[] = {
#define S(s) #s, (sizeof(#s) - 1)
	{ S(sync), do_sync },
#undef S
	{ NULL }
};

static MHD_AccessHandlerCallback
find_callback(const char **puri)
{
	const char *endpoint = *puri + 1;
	struct nssync_resource *p;
	size_t len = strcspn(endpoint, "?");

	if (len == 0)
		return cb_not_found;

	if (endpoint[len-1] == '/')
		--len;
	*puri = endpoint + len;

	for (p = nssync_resources; p->uri; p++)
		if (len >= p->uri_len
		    && memcmp(p->uri, endpoint, p->uri_len) == 0
		    && (endpoint[p->uri_len] == 0
			|| endpoint[p->uri_len] == '/'
			|| endpoint[p->uri_len] == '?')
		    && (!p->predicate || p->predicate()))
			return p->cbf;
	return cb_not_found;
}

static int
nssync_mhd_handler(void *cls,
		   struct MHD_Connection *conn,
		   const char *url, const char *method,
		   const char *version,
		   const char *upload_data, size_t *upload_data_size,
		   void **con_cls)
{
	MHD_AccessHandlerCallback cb;
	error("METHOD: %s",method);
	error("URL: %s",url);
	cb = find_callback(&url);
	return cb(cls, conn, url, method, version, upload_data,
		  upload_data_size, con_cls);
}

void
nssync_server(void)
{
	int fd;
	struct MHD_Daemon *mhd;
	
	if (!server_addr) {
		if (grecs_str_to_sockaddr(&server_addr, DEFAULT_NSSYNC_ADDR,
					  NULL, NULL))
			exit(EX_UNAVAILABLE);
	}
	
	fd = open_socket(server_addr);
	if (fd == -1)
		exit(EX_UNAVAILABLE);

	MHD_set_panic_func(nssync_mhd_panic, NULL);
	mhd = MHD_start_daemon(MHD_USE_ERROR_LOG, 0,
			       nssync_mhd_acl, NULL,
			       nssync_mhd_handler, NULL,
			       MHD_OPTION_LISTEN_SOCKET, fd,
			       MHD_OPTION_EXTERNAL_LOGGER, nssync_mhd_logger, NULL,
			       MHD_OPTION_END);
	while (1) {
		fd_set rs, ws, es;
		int max;
		int rc;
		
		FD_ZERO(&rs);
		FD_ZERO(&ws);
		FD_ZERO(&es);
		if (MHD_get_fdset(mhd, &rs, &ws, &es, &max) != MHD_YES) {
			//FIXME
			break;
		}
		rc = select(max + 1, &rs, &ws, &es, NULL);
		if (rc == -1) {
			if (errno == EINTR)
				continue;
			error("select: %s", strerror(errno));
			break;
		}
		MHD_run(mhd);
	}
	MHD_stop_daemon(mhd);
		
			
	exit(0);//FIXME
}
